//***************************************************************************
// Group Linslot / Linux - Slotrace Manager
// File iothread.cc
// Date 28.11.06 - Jrg Wendel
// This code is distributed under the terms and conditions of the
// GNU GENERAL PUBLIC LICENSE. See the file COPYING for details.
//***************************************************************************

//***************************************************************************
// Includes
//***************************************************************************


#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <iothread.hpp>
#include <common.hpp>

//***************************************************************************
// Class SlotService
//***************************************************************************
//***************************************************************************
// Is Bit
//***************************************************************************

int SlotService::isBit(unsigned char byte, int bit)
{
   int b = 1;

   for (int i = 0; i < bit; i++)   
      b = b << 1;

   return byte & b;
}

//***************************************************************************
// Toggle Bit
//***************************************************************************

void SlotService::toggleBit(unsigned char &byte, int bit)
{
   setBitTo(byte, bit, !isBit(byte, bit));
}

//***************************************************************************
// Set Bit To
//***************************************************************************

void SlotService::setBitTo(unsigned char &byte, int bit, int value)
{
   unsigned char b = 1;

   for (int i = 0; i < bit; i++)   
      b = b << 1;

   if (value)
      byte |= b;
   else
      byte &= ~b;
}

//***************************************************************************
// To Bin Str
//***************************************************************************

const char* SlotService::toBinStr(unsigned char byte)
{
   static char mask[8+1];

   int bit = 1;

   for (int i = 0; i < 8; i++)
   {
      mask[7-i] = (byte & bit) ? '1' : '0';
      
      bit = bit << 1;
   }

   mask[8] = 0;

   return mask;
}

//***************************************************************************
// kmh
//***************************************************************************

double SlotService::kmh(double meter, double usec)
{
   return ((double)meter/1000L)  /  
      ((double)usec / 1000000L / 60L / 60L);
}

//***************************************************************************
// Elapsed in Seconds
//***************************************************************************

long SlotService::elapsed(const timeval* tpFrom, const timeval* tpNow)
{
   int sec = tpNow->tv_sec - tpFrom->tv_sec;
   int usec = sec*1000000 + tpNow->tv_usec;

   usec -= tpFrom->tv_usec;

   return usec;
}

//***************************************************************************
// Seconds String
//***************************************************************************

const char* SlotService::secStr(long usec)
{
   static char buf[20];

   sprintf(buf, "%2ld.%06ld", usec/1000000, usec%1000000);

   return buf;
}

//***************************************************************************
// Class IoThread
//***************************************************************************
//***************************************************************************
// Object
//***************************************************************************

IoThread::IoThread()
   : QThread()
{
   testMode = false;
   running = false;
   devHandle = 0;
   io = 0;

   out[0] = 0xFF;
   out[1] = 0xFF;

   iow = new Iow;
}

IoThread::~IoThread()
{
   exit();

   if (iow) delete iow;
}

//***************************************************************************
// Init
//***************************************************************************

int IoThread::init(const char* dev)
{  
   // save settings

   strncpy(device, dev, 100); 
   device[100] = 0;

   // Open device
   
   if (open() != 0)
      return -1;

   // Init all inputs  to 1
   //      all outputs to 0

   writeByte(io, 0, 0xFF);
   writeByte(io, 1, 0x00);  

   return 0;
}

//***************************************************************************
// Exit
//***************************************************************************

int IoThread::exit()
{
   // on exit set all I/O ports back to 1

   if (io)
   {
      writeByte(io, 0, 0xFF);
      writeByte(io, 1, 0xFF);
   }

   // Close device
   
   close();

   return 0;
}

//***************************************************************************
// Open
//***************************************************************************

int IoThread::open()
{
   int numIows;

   if (devHandle)
      close();

   devHandle = iow->OpenDevice(device);
   
   if (devHandle == 0)
   {
      tell(eloAlways, "Failed to open device '%s<N>'", device);
      return -1;
   }

   // Get number of IOWs in system

   numIows = iow->GetNumDevs();
   tell(eloDetail, "%d IOWs in system", numIows);

   if (!numIows)
   {
      tell(eloAlways, "No IOWarrior found, aborting!");
      return -1;
   }

   io = iow->GetDeviceHandle(1);

   iow->SetWriteTimeout(io, 1000);
   iow->SetTimeout(io, 1000);

   return 0;
}

//***************************************************************************
// Close
//***************************************************************************

int IoThread::close()
{
   if (devHandle)
   {
      iow->CloseDevice(devHandle);
      devHandle = 0;
   }
   
   return 0;
}

//***************************************************************************
// Reopen
//***************************************************************************

int IoThread::reopen()
{
   close();

   devHandle = iow->OpenDevice(device);
   
   if (devHandle == 0)
   {
      tell(eloAlways, "Failed to open device '%s<N>'", device);
      return -1;
   }

   // Init all inputs  to 1
   //      all outputs to last value!

   writeByte(io, 0, 0xFF);       // Inputs always FF
   writeByte(io, 1, out[1]);     // last output value

   return 0;
}

//***************************************************************************
// Flush Input Stack
//***************************************************************************

int IoThread::flushStack(Iow::IowHandle io)
{
   unsigned int tmp;

   if (!io)
      return 0;

   // flush input stack

   // QMutexLocker locker(&mutex);

   while (iow->ReadImmediate(io, &tmp))
      ;

   return 0;
}

//***************************************************************************
// Look Byte
//***************************************************************************

int IoThread::lookByte(Iow::IowHandle io, int byteNum, unsigned char* byte)
{
   IowValue value;

   if (!io)
      return false;

   // QMutexLocker locker(&mutex);

   if (iow->ReadImmediate(io, &value.dword))
   {
      *byte = value.bytes[byteNum];

      return true;
   }

   return false;
}

//***************************************************************************
// Wait Byte
//***************************************************************************

int IoThread::waitByte(Iow::IowHandle io, int byteNum, unsigned char* byte)
{
   Iow::Report24 rep;
   int ret;

   if (!io)
      usleep(500000);
      
   if (testMode)
   {
      static unsigned char last = 0x00;
      last = ~last;
      *byte = last;
      return true;
   }

   if (!io)
      return false;

   // QMutexLocker locker(&mutex);

   ret = iow->Read(io, (char*)&rep, sizeof(Iow::Report24));

   if (ret == sizeof(Iow::Report24))
   {
      *byte = rep.Bytes[byteNum];
      
      return true;
   }
   else if (ret < 0)
   {
      reopen();
   }

   return false;
}

//***************************************************************************
// Write Byte
//***************************************************************************

int IoThread::writeByte(Iow::IowHandle devHandle, int byteNum, unsigned char byte)
{
   Iow::Report rep;
   int res = false;

   // Init report

   rep.ReportID = 0;
   rep.Value = 0xFFFF;
   rep.Bytes[byteNum] = byte;
   out[byteNum] = byte;

   unsigned char old;

   tell(eloDebug, "Debug: Writing to byte %d", byteNum);

   if (byteNum == 0)
   {
      if (lookByte(devHandle, 1, &old))
         rep.Bytes[1] = old;

      out[1] = old;
   }
   else
   {
      if (lookByte(devHandle, 0, &old))
         rep.Bytes[0] = old;

      out[0] = old;
   }

 
   if (io)
   {
      // QMutexLocker locker(&mutex);

      tell(eloDebug, "Debug: Writing '%s' to port %d", toBinStr(rep.Bytes[0]), 0);
      tell(eloDebug, "Debug: Writing '%s' to port %d", toBinStr(rep.Bytes[1]), 1);

      res = iow->Write(devHandle, (char*)&rep, sizeof(Iow::Report)) 
         == sizeof(Iow::Report);
   }
   
   return res;
}

//***************************************************************************
// Write Bit
//***************************************************************************

void IoThread::writeBit(int bit, int value)
{
   unsigned char byte = bit < 8 ? out[0] : out[1];

   tell(eloDebug, "Debug: Write bit (%d) value (%d)", bit, value);

   setBitTo(byte, bit < 8 ? bit : bit-8, value);

   writeByte(io, bit < 8 ? 0 : 1, byte);
}

//***************************************************************************
// Read Out Bit
//***************************************************************************

int IoThread::readBit(int bit)
{
   unsigned char byte = bit < 8 ? out[0] : out[1];
   
   tell(eloDebug, "Debug: Byte is '%s' on port %d", toBinStr(byte), bit < 8 ? 0 : 1);

   return isBit(byte, bit < 8 ? bit : bit-8);
}

//***************************************************************************
// Run
//***************************************************************************

void IoThread::run()
{
   IoEvent ioEvent;

   // first flush old input data
   
   flushStack(io);

   running = true;

   // loop

   while (running)
   {
      if (waitByte(io, 0, &ioEvent.byte))
      {
         // Zeit merken
         
         gettimeofday(&ioEvent.tp, 0);

         tell(eloDebug, "Debug: Input signal detected");

         emit onIoSignal(ioEvent);
      }
   }
}
