//***************************************************************************
// Group Linslot / Linux - Slotrace Manager
// File iow.cc
// Date 05.11.06 - Jrg Wendel
// This code is distributed under the terms and conditions of the
// GNU GENERAL PUBLIC LICENSE. See the file COPYING for details.
//***************************************************************************

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <asm/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <linux/input.h>
#include <linux/ioctl.h>

#include "iow.hpp"

//***************************************************************************
// Object
//***************************************************************************

Iow::Iow()
{
   numIoWarriors = 0;
}

//***************************************************************************
// Clear
//***************************************************************************

void Iow::Clear()
{
   unsigned int i;

   for (i = 0; i < numIoWarriors; i++)
   {
      if (IoWarriors[i].fdIO != -1)
         close(IoWarriors[i].fdIO);
   }

   memset(IoWarriors, 0, sizeof(IoWarriors));
   numIoWarriors = 0;
}

//***************************************************************************
// Get Device By Handle
//***************************************************************************

Iow::IowDevice* Iow::GetDeviceByHandle(IowHandle iowHandle)
{
   unsigned int i;

   for (i = 0; i < numIoWarriors; i++)
      if ((IowDevice *) iowHandle == &IoWarriors[i])
         return &IoWarriors[i];

   return 0;
}

//***************************************************************************
// Open Device
//***************************************************************************

Iow::IowHandle Iow::OpenDevice(const char* dev)
{
   char ifname[80];         // The filename for the io interface
   iowarrior_info info;     // the info for the device
   int i;
   unsigned int j;
   int fd;
   int ioctlret;

   Clear();

   // First step : open the interface 0 of all IOWarriors found

   for (i = 0; i < maxDevices*2; i++)
   {
      sprintf(ifname, "%s%d", dev, i);
      fd = open(ifname, O_RDWR);

      if (fd != -1)
      {
         // now get the info for this device

         ioctlret = ioctl(fd, IOW_GETINFO, &info);

         if (ioctlret != -1 && info.if_num == 0)
         {
            IoWarriors[numIoWarriors].dev_num = i;
            IoWarriors[numIoWarriors].info = info;
            IoWarriors[numIoWarriors].fdIO = fd;
            IoWarriors[numIoWarriors].lastValue = 0xFFFFFFFF;
            IoWarriors[numIoWarriors].readTimeoutVal = 0xFFFFFFFF;
            IoWarriors[numIoWarriors].writeTimeoutVal = 0xFFFFFFFF;
            numIoWarriors++;
         }
         else
            close(fd);
      }
   }

   // now look for interface1

   for (i = 0; i < maxDevices*2; i++)
   {
      sprintf(ifname, "%s%d", dev, i);
      fd = open(ifname, O_RDWR);

      if (fd != -1)
      {
         // now get the info for this device

         ioctlret = ioctl(fd, IOW_GETINFO, &info);

         if (ioctlret != -1 && info.if_num == 1)
         {
            for (j = 0; j < numIoWarriors; j++)
            {
               if (IoWarriors[j].info.product == info.product)
               {
                  if (info.serial[0] == '\0' ?
                     i - 1 == IoWarriors[j].dev_num : strcmp(IoWarriors[j].info.serial, info.serial) == 0)
                  {
                     fd = -1;
                     break;
                  }
               }
            }
         }
      }

      if (fd != -1)
         close(fd);
   }

   return GetDeviceHandle(1);
}

//***************************************************************************
// Get Device Handle
//***************************************************************************

Iow::IowHandle Iow::GetDeviceHandle(unsigned int numDevice)
{
   if (numDevice >= 1 && numDevice <= numIoWarriors)
      return (IowHandle) &IoWarriors[numDevice - 1];
   else
      return 0;
}

//***************************************************************************
// Close Device
//***************************************************************************

void Iow::CloseDevice(IowHandle /*devHandle*/)
{ 
   Clear();
}

//***************************************************************************
// Set Timeout
//***************************************************************************

bool Iow::SetTimeout(IowHandle devHandle, unsigned int timeout)
{
   IowDevice* dev;

   dev = GetDeviceByHandle(devHandle);

   if (dev)
   {
      dev->readTimeout.tv_sec = (time_t)(timeout / 1000);
      dev->readTimeout.tv_usec = (long) ((timeout % 1000) * 1000);
      dev->readTimeoutVal = timeout;
   }

   return dev != 0;
}

//***************************************************************************
// Set Write Timeout
//***************************************************************************

bool Iow::SetWriteTimeout(IowHandle devHandle, unsigned int timeout)
{
   IowDevice* dev;

   dev = GetDeviceByHandle(devHandle);

   if (dev)
   {
      dev->writeTimeout.tv_sec = (time_t)(timeout / 1000);
      dev->writeTimeout.tv_usec = (long)((timeout % 1000) * 1000);
      dev->writeTimeoutVal = timeout;
   }

   return dev != 0;
}

//***************************************************************************
// Read
//***************************************************************************

int Iow::Read(IowHandle devHandle, char* buffer, int length)
{
   IowDevice *dev;
   timeval tv;
   fd_set rfds;
   int fd;
   int count;
   int result, size, extra;

   dev = GetDeviceByHandle(devHandle);

   if (!dev || !buffer)
      return -1;

   fd = dev->fdIO;
   size = dev->info.packet_size;
   extra = 1;
      
   for (count = 0; count + size + extra <= length; )
   {
      tv = dev->readTimeout;
      FD_ZERO(&rfds);      // clear the read-descriptor for select
      FD_SET(fd, &rfds);   // add the filedescriptor to read-descriptor

      if (dev->readTimeoutVal != 0xFFFFFFFF)
         result = select(fd + 1, &rfds, 0, 0, &tv);
      else
         result = select(fd + 1, &rfds, 0, 0, 0);

      if (result > 0 && FD_ISSET(fd, &rfds))
      {
         int res;

         *buffer = '\0';
         buffer += extra;

         // there should be some data ready for reading now

         res = read(fd, buffer, size);

         if (res < 0)
         {
            printf("Error: Read failed, system error was '%s' (%d)\n", 
                   strerror(errno), errno);

            return -1;
         }
         else if (res != size)
            break;

         buffer += size;
         count += size + extra;
      }

      else
      {
         // if we encountered a timeout then do not try to read several reports

         if (!result)
            break;
      }
   }

   return count;
}

//***************************************************************************
// Read Immediate
//***************************************************************************

bool Iow::ReadImmediate(IowHandle devHandle, unsigned int* value)
{
   IowDevice *dev;
   timeval tv;
   fd_set rfds;
   int result;
   Report report;

   dev = GetDeviceByHandle(devHandle);

   if (dev)
   {
      if (!value)
         return 0;

      tv.tv_sec = 0;
      tv.tv_usec = 0;
      FD_ZERO(&rfds);            // clear the read-descriptor for select
      FD_SET(dev->fdIO, &rfds);  // add the filedescriptor to read-descriptor
      result = select(dev->fdIO + 1, &rfds, 0, 0, &tv);

      if (result > 0 && FD_ISSET(dev->fdIO, &rfds))
      {
         memset(&report, 0, sizeof(report));

         // there should be some data ready for reading now

         if (read(dev->fdIO, &report.Value, dev->info.packet_size) == (int)dev->info.packet_size)
         {
            dev->lastValue = report.Value;
            *value = report.Value;

            return true;
         }
      }
   }

   *value = dev->lastValue;

   return false;
}

//***************************************************************************
// Write
//***************************************************************************

int Iow::Write(IowHandle devHandle, char* buffer, int length)
{
   IowDevice *dev;
   int fd;
   int count;
   int siz, extra;

   dev = GetDeviceByHandle(devHandle);

   if (!buffer)
      return 0;

   if (!dev)
      return -1;

   fd = dev->fdIO;
   siz = dev->info.packet_size;
   extra = 1;
   
   for (count = 0; count + siz + extra <= length; count += siz + extra)
   {
      buffer += extra;
      
      if (ioctl(fd, IOW_WRITE, buffer) != 0)
         break;
      
      buffer += siz;
   }
   
   return count;
}
