//***************************************************************************
// Group Linslot / Linux - Slotrace Manager
// File linslot.cc
// Date 05.12.06 - Jrg Wendel
// This code is distributed under the terms and conditions of the
// GNU GENERAL PUBLIC LICENSE. See the file COPYING for details.
//***************************************************************************

//***************************************************************************
// Include
//***************************************************************************

#include <QMessageBox>
#include <QMetaType>
#include <QSound>

#include <sys/time.h>

#include <sqlite.hpp>
#include <linslot.hpp>

//***************************************************************************
// Local Helpers
//***************************************************************************
//***************************************************************************
// To Str
//***************************************************************************

const char* toStr(int value)
{
   static char buf[10];

   sprintf(buf, "%d", value);

   return buf;
}

//***************************************************************************
// Linslot Window
//***************************************************************************
//***************************************************************************
// Object
//***************************************************************************

LinslotWindow::LinslotWindow()
   : QMainWindow()
{
   // first init class private

   testMode = false;
   theEloquence = eloAlways;
   raceRunning = false;
   countdownStarted = false;
   countdown = 0;
   fastLapTime = 0;
   resourcePath = "/usr/local/share/linslot/";
   configPath = QString(QDir::homePath() + "/.linslot");

   QDir d(configPath);

   if (!d.exists())
      d.mkdir(configPath);

   for (int i = 0; i < slotCount; i++)
   {
      theSlots[i].round = -1;
      theSlots[i].penalty = 0;
      *theSlots[i].driver = 0;
   }

   // now setup the user interface

   setupUi(this);
   
   // dialogs

   setupDialog = new SetupDialog(configPath);
   highscoreDialog = new HighscoreDialog();
   // setupDialog->setWindowFlags(setupDialog->windowFlags() | Qt::MSWindowsFixedSizeDialogHint);

   connect(setupDialog, SIGNAL(accepted()),
           this, SLOT(onOptionsAccepted()));

   onOptionsAccepted();

   // init

   init();

   // init timer

   timer = new QTimer(this);
   connect(timer, SIGNAL(timeout()), this, SLOT(onTimer()));

   timerElapsed = new QTimer(this);
   connect(timerElapsed, SIGNAL(timeout()), this, SLOT(onElapsedTimer()));

   timerPenalty = new QTimer(this);
   connect(timerPenalty, SIGNAL(timeout()), this, SLOT(onPenaltyTimer()));

   // thread stuff 

   thread = new IoThread();

   thread->init(setupDialog->getUsbDevice());
   thread->start();
   thread->setTestMode(testMode);

   // connect thread

   qRegisterMetaType<IoEvent>("IoEvent");

   connect(thread, SIGNAL(onIoSignal(const IoEvent &)),
           this, SLOT(onIoSignal(const IoEvent &)));

   //

   setSlotPower(psAll, false);
}

LinslotWindow::~LinslotWindow()
{
   exit();
   
   delete highscoreDialog;
   delete setupDialog;
   delete db;
   delete thread;
}

//***************************************************************************
// init/exit
//***************************************************************************

void LinslotWindow::init()
{
   int i;

   // get setup parameters

   setupDialog->getSettings()->beginGroup("common");

   if ((i = comboBoxDriver1->findText(setupDialog->getSettings()->value("driver0", "").toString())) >= 0)
      comboBoxDriver1->setCurrentIndex(i);

   if ((i = comboBoxDriver2->findText(setupDialog->getSettings()->value("driver1", "").toString())) >= 0)
      comboBoxDriver2->setCurrentIndex(i);

   setupDialog->getSettings()->endGroup();

   // widged stuff

   ledsParent = labelLed00->parent();           // !!
   tableWidgetSlot1->setColumnWidth(0, 125);
   tableWidgetSlot1->setColumnWidth(1, 130);
   tableWidgetSlot2->setColumnWidth(0, 125);
   tableWidgetSlot2->setColumnWidth(1, 130);
   tableWidgetSlot1->setSelectionBehavior(QAbstractItemView::SelectRows);
   tableWidgetSlot2->setSelectionBehavior(QAbstractItemView::SelectRows);
   tableWidgetSlot1->setEditTriggers(QAbstractItemView::NoEditTriggers);
   tableWidgetSlot2->setEditTriggers(QAbstractItemView::NoEditTriggers);

   // parse arguments ..

   for (int i = 0; i < QCoreApplication::arguments().size(); i++)
   {
      const char* arg = QCoreApplication::arguments().at(i).toAscii();

      if (arg[0] != '-' || strlen(arg) != 2)  continue;

      switch (arg[1])
      {
         case 't':
         {
            testMode = true;
            break;
         }
         case 'e':
         {
            theEloquence = QCoreApplication::arguments().at(i).toInt();
            i++;
            break;
         }
      }
   }

   //

   resetWidgets();
   setLeds(lsAllOff | lsYellow1 | lsYellow2);

   // db

   if (openDb() != 0)
   {
      pushButtonSaveToDb->setEnabled(false);
      pushButtonHallOfFame->setEnabled(false);
   }

   // at last, check sound

   if (!QSound::isAvailable())
      tell(eloAlways, "No sound available, missing \"Network audio System (NAS)\"");
}

void LinslotWindow::exit()
{
   // db

   if (db) db->close();

   // save config

   setupDialog->getSettings()->beginGroup("common");
   setupDialog->getSettings()->setValue("driver0", theSlots[0].driver);
   setupDialog->getSettings()->setValue("driver1", theSlots[1].driver);
   setupDialog->getSettings()->endGroup();

   // stop thread

   tell(eloAlways, "Stopping thread");
   thread->stop();

   if (!thread->wait(5000))
      tell(eloAlways, "Thread would not end!");
   else 
      tell(eloAlways, "Thread ended regularly");
}

//***************************************************************************
// Reset Widgets
//***************************************************************************

void LinslotWindow::resetWidgets()
{
   pushButtonInitUsb->setVisible(false);

   tableWidgetSlot1->setEnabled(false);
   tableWidgetSlot2->setEnabled(false);
   tableWidgetSlot1->clear();
   tableWidgetSlot2->clear();
   tableWidgetSlot1->setEnabled(true);
   tableWidgetSlot2->setEnabled(true);

   QStringList header;
   header << "Zeit" << "km/h";
   tableWidgetSlot1->setHorizontalHeaderLabels(header);
   tableWidgetSlot2->setHorizontalHeaderLabels(header);

   labelInfo->setText("Standby");
   labelFastLap->setText("");

   tableWidgetSlot1->setRowCount(setupDialog->getRoundCount());
   tableWidgetSlot2->setRowCount(setupDialog->getRoundCount());
   labelRoundCounter1->setText(toStr(setupDialog->getRoundCount()));
   labelRoundCounter2->setText(toStr(setupDialog->getRoundCount()));
}

//***************************************************************************
// Get Images For
//***************************************************************************

int LinslotWindow::getImagesFor(int led, const char* &ledOn, const char* &ledOff)
{
   switch (led)
   {
      case lsRed1:
      case lsRed2:
      case lsRed3:
      case lsRed4:
      case lsRed5: 
         ledOn  = "/pixmap/led-red-on.gif";
         ledOff = "/pixmap/led-red-off.gif";
         break;
      case lsGreen:
         ledOn  = "/pixmap/led-green-on.gif";
         ledOff = "/pixmap/led-green-off.gif";
         break;
      case lsYellow1:
      case lsYellow2:
         ledOn  = "/pixmap/led-yellow-on.gif";
         ledOff = "/pixmap/led-yellow-off.gif";
         break;

      default: return -1;
   }

   return 0;
}

//***************************************************************************
// Set Leds
//***************************************************************************

void LinslotWindow::setLeds(int mask)
{
   const char* ledOn  = "";
   const char* ledOff = "";   

   for (int i = 0; i < ledsParent->children().size(); i++)
   {
      QLabel* label = qobject_cast<QLabel *>(ledsParent->children().at(i));

      if (label)
      {
         if (getImagesFor(label->indent(), ledOn, ledOff) != 0)
            continue;

         if (mask & label->indent())
            label->setPixmap(QPixmap(QString(resourcePath) + ledOn));
         else
            label->setPixmap(QPixmap(QString(resourcePath) + ledOff));
      }
   }
}

//***************************************************************************
// Switch Leds
//***************************************************************************

void LinslotWindow::switchLeds(int mask, int state)
{
   const char* ledOn  = "";
   const char* ledOff = "";   

   for (int i = 0; i < ledsParent->children().size(); i++)
   {
      QLabel* label = qobject_cast<QLabel *>(ledsParent->children().at(i));

      if (label && mask & label->indent())
      {
         if (getImagesFor(label->indent(), ledOn, ledOff) != 0)
            continue;

         if (state)
            label->setPixmap(QPixmap(QString(resourcePath) + ledOn));
         else
            label->setPixmap(QPixmap(QString(resourcePath) + ledOff));
      }
   }
}

//***************************************************************************
// Set Slot Power
//***************************************************************************

void LinslotWindow::setSlotPower(int slot, int flag)
{
   tell(eloAlways, "set slot power to %d", flag);

   if (slot & psSlot1)
   {
      switchLeds(lsYellow1, !flag); 
      thread->writeBit(bitPowerSlot1, flag);
   }

   if (slot & psSlot2)
   {
      switchLeds(lsYellow2, !flag);
      thread->writeBit(bitPowerSlot2, flag);
   }
}

//***************************************************************************
// On Options Accepted
//***************************************************************************

void LinslotWindow::onOptionsAccepted()
{
   QListWidgetItem* listItem;

   for (int i = 0; i < setupDialog->listWidgetDriver->count(); i++)
   {
      if ((listItem = setupDialog->listWidgetDriver->item(i)))
      {
         if (comboBoxDriver1->findText(listItem->text()) == -1)
         {
            comboBoxDriver1->addItem(listItem->text());
            comboBoxDriver2->addItem(listItem->text());
         }
      }
   }

   labelRoundCounter1->setText(toStr(setupDialog->getRoundCount()));
   labelRoundCounter2->setText(toStr(setupDialog->getRoundCount()));
   tableWidgetSlot1->setRowCount(setupDialog->getRoundCount());
   tableWidgetSlot2->setRowCount(setupDialog->getRoundCount());
}

//***************************************************************************
// On Driver1 Index Changed
//***************************************************************************

void LinslotWindow::on_comboBoxDriver1_currentIndexChanged(QString value)
{
   strncpy(theSlots[0].driver, value.toAscii(), sizeName);
}

//***************************************************************************
// On Driver2 Index Changed
//***************************************************************************

void LinslotWindow::on_comboBoxDriver2_currentIndexChanged(QString value)
{
   strncpy(theSlots[1].driver, value.toAscii(), sizeName);
}

//***************************************************************************
// On Penalty Timer
//***************************************************************************

void LinslotWindow::onPenaltyTimer()
{
   int n = 0;

   tell(eloAlways, "onPenaltyTimer");

   for (int i = 0; i < slotCount; i++)
   {
      if (theSlots[i].penalty > 0)
      {
         theSlots[i].penalty--;

         if (theSlots[i].penalty <= 0)
            setSlotPower(i+1, true);
         else
            n += theSlots[i].penalty;
      }
   }

   if (n <= 0)
      timerPenalty->stop();
}

//***************************************************************************
// On Elapsed Timer
//***************************************************************************

void LinslotWindow::onElapsedTimer()
{
   timeval tp;
   int finished = false;

   gettimeofday(&tp, 0);
   
   int usec = elapsed(&raceStart, &tp);

   char* buf;
   asprintf(&buf, "%4.1f", (double)usec/1000000L);
   labelElapsed->setText(buf);
   free(buf);

   if (radioButtonRoundRace->isChecked() && setupDialog->getMaxTimeRoundRace())
      finished = usec/1000000 >= setupDialog->getMaxTimeRoundRace();
   else if (radioButtonTraining->isChecked() && setupDialog->getMaxTimeTraining())
      finished = usec/1000000 >= setupDialog->getMaxTimeTraining();
   
   if (finished)
   {
      int s = theSlots[0].round == theSlots[1].round ? -1 : 
         theSlots[0].round > theSlots[1].round ? 0 : 1;

      tell(eloDetail, "Finished after %d seconds of race/training", usec/1000000);
      atFinish(!radioButtonTraining->isChecked() ? s : -1);
   }
}

//***************************************************************************
// On Timer
//***************************************************************************

void LinslotWindow::onTimer()
{
   if (countdown == 5)
   {
      switchLeds(lsPhase5, false);
      switchLeds(lsGreen, true);
      atStart();
   }
   else
   {
      QSound::play(QString(resourcePath)+ "/sound/beep.wav");

      switch (countdown)
      {
         case 0: switchLeds(lsPhase1, true); break;
         case 1: switchLeds(lsPhase2, true); break;
         case 2: switchLeds(lsPhase3, true); break;
         case 3: switchLeds(lsPhase4, true); break;
         case 4: switchLeds(lsPhase5, true); break;
      }
      
      countdown++;
   }
}

//***************************************************************************
// On I/O Signal
//***************************************************************************

void LinslotWindow::onIoSignal(const IoEvent &ioEvent)
{
   IoEvent theEvent = ioEvent;

   // entprellen

   if (bounceCheck(theEvent))
      return ;

   if (!testMode)
   {
      // externer Start/Stop Taster

      if (!isBit(theEvent.byte, 2))
      {
         tell(eloDebug, "Signal bit 2");

         if (!countdownStarted)
            atStartCountdown();
         else
            atAbort("Abbruch");
      }
   }
   else
   {
      if (!raceRunning)
         return ;
   }
   
   // process slot signals
   
   if (!isBit(theEvent.byte, 0))
      atSlotSignal(0, &theEvent.tp);
   
   if (!raceRunning && !countdownStarted)
      return ;
   
   if (!isBit(theEvent.byte, 1))
      atSlotSignal(1, &theEvent.tp);
}

//***************************************************************************
// On Button Power
//***************************************************************************

void LinslotWindow::on_pushButtonPower_clicked()
{ 
   setSlotPower(psAll, !thread->readBit(bitPowerSlot1));
}

//***************************************************************************
// On Button Start
//***************************************************************************

void LinslotWindow::on_pushButtonStartRace_clicked()
{
   if (!countdownStarted)
      atStartCountdown();
   else
      atAbort("Abbruch");
}

//***************************************************************************
// On Button Options
//***************************************************************************

void LinslotWindow::on_pushButtonOptions_clicked()
{
   setupDialog->show();
}

//***************************************************************************
// On Button Hall Of Fame
//***************************************************************************

void LinslotWindow::on_pushButtonHallOfFame_clicked()
{
   if (!db)
      return ;

   highscoreDialog->setDb(db);
   highscoreDialog->fill();
   highscoreDialog->show();
}

//***************************************************************************
// On Button Save To Db
//***************************************************************************

void LinslotWindow::on_pushButtonSaveToDb_clicked()
{
   saveRace();
}

//***************************************************************************
// At Start
//***************************************************************************

void LinslotWindow::atStartCountdown()
{
   countdownStarted = true;
   countdown = 0;

   // init slot data

   for (int i = 0; i < slotCount; i++)
   {
      theSlots[i].penalty = 0;
      theSlots[i].round = -1;
   }

   // power on

   setSlotPower(psAll, true);

   // set/reset widget stuff

   radioButtonTraining->setEnabled(radioButtonTraining->isChecked());
   radioButtonRoundRace->setEnabled(radioButtonRoundRace->isChecked());
   pushButtonOptions->setEnabled(false);

   resetWidgets();
   labelInfo->setText("Countdown");
   pushButtonStartRace->setText("&Abbrechen");
   labelElapsed->setText("0.0");
   timer->start(1000);
}

//***************************************************************************
// At Start
//***************************************************************************

void LinslotWindow::atStart()
{ 
   // init

   countdown = 0;
   fastLapTime = 0;
   raceRunning = true;
   gettimeofday(&raceStart, 0);
   timer->stop();
   timerElapsed->start(100);          // 0.1 sec
   timerPenalty->start(1000);         // 1.0 sec

   for (int i = 0; i < slotCount; i++)
      theSlots[i].lastSignal = raceStart;

   if (radioButtonRoundRace->isChecked())
      labelInfo->setText("Rennen luft");
   else
      labelInfo->setText("Freies Training");

   QSound::play(QString(resourcePath)+ "/sound/start.wav");
}

//***************************************************************************
// At Finish
//***************************************************************************

void LinslotWindow::atFinish(int slot)
{   
   timeval tp;
   int usec;
   char* buf;

   gettimeofday(&tp, 0);

   if (slot >= 0)
      asprintf(&buf, "! ENDE ! Sieger %s", theSlots[slot].driver);
   else
      asprintf(&buf, "! ENDE !");

   atStop(buf);
   free(buf);

   QSound::play(QString(resourcePath)+ "/sound/finish.wav");

   usec = elapsed(&raceStart, &tp);

   tell(eloDebug, "Total: %2.2d:%2.2d.%06d",
        usec/1000000/60, usec/1000000%60, usec%1000000);
  
   for (int i = 0; i < slotCount; i++)
      tell(eloDebug, "Bahn %d - %d Runden gefahren", i+1, theSlots[i].round);
}

//***************************************************************************
// At Abort
//***************************************************************************

void LinslotWindow::atAbort(const char* info)
{
   atStop(info);

   for (int i = 0; i < 5; i++, usleep(100000))
      QSound::play(QString(resourcePath)+ "/sound/beep.wav");
}

//***************************************************************************
// At Stop
//***************************************************************************

void LinslotWindow::atStop(const char* info)
{
   timeval tp;
   char* buf;
   int usec;

   raceRunning = false;
   countdownStarted = false;
   timer->stop();
   timerElapsed->stop();
   
   // slot power off

   setLeds(lsAllOff | lsYellow1 | lsYellow2);
   setSlotPower(psAll, false);

   // update widgets

   gettimeofday(&tp, 0);
   usec = elapsed(&raceStart, &tp);
   asprintf(&buf, "%4.1f", (double)usec/1000000L);
   labelElapsed->setText(buf);
   free(buf);

   pushButtonStartRace->setText("St&arten");

   labelInfo->setText(info);

   radioButtonTraining->setEnabled(true);
   radioButtonRoundRace->setEnabled(true);
   pushButtonOptions->setEnabled(true);
}

//***************************************************************************
// At Jump The Gun
//***************************************************************************
  
void LinslotWindow::atJumpTheGun(int slot)
{
   char* info;

   if (setupDialog->getAbortAtJumpTheGun())
   {
      asprintf(&info, "Abbruch Frhstart %s", theSlots[slot].driver);
      atAbort(info);
      free(info);
   }
   else
   {
      setSlotPower(slot+1, false);
      theSlots[slot].penalty += setupDialog->getPenaltyAtJumpTheGun();
   }
}

//***************************************************************************
// Bounce Check  (entprellen)
//***************************************************************************

int LinslotWindow::bounceCheck(IoEvent &ioEvent)
{
   static int initialized = false;
   static timeval last[bitsPerByte] = { { 0, 0 } };

   int i;

   // init 

   if (!initialized)
   {
      initialized = true;

      for (i = 0; i < bitsPerByte; i++)
      {
         gettimeofday(&last[i], 0);
         last[i].tv_sec -= 10;
      }
   }

   // input signal pending ?

   if (ioEvent.byte == 0xFF)
      return true;

   for (i = 0; i < bitsPerByte; i++)
   {
      if (!isBit(ioEvent.byte, i))
      {
         // if < bounceTime remove change

         if (elapsed(&last[i], &ioEvent.tp) < (i < 2 ? bounceTimeSlot : bounceTime))
            toggleBit(ioEvent.byte, i);
         else 
            last[i] = ioEvent.tp;
      }
   }

   if (ioEvent.byte == 0xFF)
      return true;

   return false;
}

//***************************************************************************
// At Slot Signal
//***************************************************************************

void LinslotWindow::atSlotSignal(int slot, const timeval* tp)
{
   char* tmp;
   int rnd = 0;

   tell(eloDebug, "Debug: Signal slot %d", slot);
   QSound::play(QString(resourcePath)+ "/sound/beep.wav");

   theSlots[slot].round++;

   // 'to jump the gun' (Fruehstart) ?

   if (!raceRunning && !testMode)   
      atJumpTheGun(slot);
   
   if (!theSlots[slot].round)
      return ;

   // calc

   int usec = thread->elapsed(&theSlots[slot].lastSignal, tp);

   tell(eloDetail, "%d:%d)  %2d.%03d Sekunden (%02.2f km/h)",
        slot, theSlots[slot].round,
        usec/1000000, usec%1000000,
        kmh(setupDialog->getSlotLength(), usec));

   if (radioButtonTraining->isChecked() && 
       tableWidgetSlot1->rowCount() <= theSlots[slot].round)
   {
      tableWidgetSlot1->setRowCount(theSlots[slot].round+10);
      tableWidgetSlot2->setRowCount(tableWidgetSlot1->rowCount());
   }

   asprintf(&tmp, " %2.3f", ((double)usec)/1000000L);
   QTableWidgetItem* itemTime = new QTableWidgetItem(tmp);
   free(tmp);

   asprintf(&tmp, "%02.2f",  kmh(setupDialog->getSlotLength(), usec)*setupDialog->getSpeedFactor());
   QTableWidgetItem* itemKmh = new QTableWidgetItem(tmp);
   free(tmp);

   if (!fastLapTime || fastLapTime > usec)
   {
      fastLapTime = usec;

      asprintf(&tmp, "Schnellste Runde\n%s  - %2.3fs (%02.2fkm/h)", 
               *theSlots[slot].driver ? theSlots[slot].driver : "",
               ((double)usec)/1000000L, 
               kmh(setupDialog->getSlotLength(), usec)*setupDialog->getSpeedFactor());
      labelFastLap->setText(tmp);
      free(tmp);
   }

   if (radioButtonRoundRace->isChecked())
      rnd = setupDialog->getRoundCount()-theSlots[slot].round;
   else
      rnd = theSlots[slot].round;

   if (slot == 0)
   {
      labelRoundCounter1->setText(toStr(rnd));
      tableWidgetSlot1->setItem(theSlots[slot].round-1, 0, itemTime);
      tableWidgetSlot1->setItem(theSlots[slot].round-1, 1, itemKmh);
      tableWidgetSlot1->setCurrentCell(theSlots[slot].round-1, 0);
   }
   else
   {
      labelRoundCounter2->setText(toStr(rnd));
      tableWidgetSlot2->setItem(theSlots[slot].round-1, 0, itemTime);
      tableWidgetSlot2->setItem(theSlots[slot].round-1, 1, itemKmh);
      tableWidgetSlot2->setCurrentCell(theSlots[slot].round-1, 0);
   }

   gettimeofday(&theSlots[slot].lastSignal, 0);

   if (radioButtonRoundRace->isChecked() && 
       theSlots[slot].round >= setupDialog->getRoundCount())
   {
      tell(eloDetail, "Race finished after %d laps of slot %d", theSlots[slot].round, slot);
      atFinish(slot);
   }
   else if (radioButtonTraining->isChecked() && 
       theSlots[slot].round >= setupDialog->getRoundLimit())
   {
      tell(eloDetail, "Training finished after %d laps of slot %d", theSlots[slot].round, slot);
      atFinish(-1);
   }
}

//***************************************************************************
// Create Db
//***************************************************************************

int LinslotWindow::openDb()
{
   char* dbPath;
   int status;

   asprintf(&dbPath, "%s/%s", configPath.toAscii().constData(), 
            setupDialog->getDatabaseName());

   db = new SqliteDb(dbPath);

   if ((status = db->open()) != 0)
   {
      tell(eloAlways, "Error: Open database '%s' failed. "  \
           "Result was (%d)", dbPath, status);
   }

   else
   {
      if ((status = db->execute("select * from races;")) != 0)
      {
         if (QMessageBox::question(this, "Warnung", "Tabellen nicht gefunden, anlegen?", 
                                   QMessageBox::Yes, QMessageBox::No) == QMessageBox::Yes)
         {
            createDb();
            
            if ((status = db->execute("select * from races;")) != 0)
               QMessageBox::warning(this, "Fehler", "Tabellen konnten nicht angelegt werden!");
         }
      }
   }
   
   if (status != 0)
   {
      db->close();
      delete db;
      db = 0;
   }

   return status;
}

//***************************************************************************
// Create Db
//***************************************************************************

int LinslotWindow::createDb()
{
   tell(eloDebug, "Creating tables");

   db->execute("CREATE TABLE races ("                          \
               "RACE_ID INTEGER PRIMARY KEY AUTOINCREMENT,"    \
               "DRIVER1 INTEGER, "                             \
               "DRIVER2 INTEGER, "                             \
               "DATE DATETIME, "                               \
               "LAPS INTEGER, "                                \
               "LAP_LENGTH REAL, "                             \
               "COURSE TEXT"                                   \
               ");");
   
   db->execute("CREATE TABLE laps ("                        \
               "LAP_ID INTEGER PRIMARY KEY AUTOINCREMENT,"  \
               "RACE_ID INTEGER, "                          \
               "DRIVER_NR INTEGER, "                        \
               "LAP_NR INTEGER, "                           \
               "LAP_TIME REAL"                              \
               ");");
   
   db->execute("CREATE TABLE drivers ("                        \
               "DRIVER_ID INTEGER PRIMARY KEY AUTOINCREMENT,"  \
               "NAME TEXT"                                     \
               ");");

   db->execute("CREATE TABLE courses ("                        \
               "COURSE_ID INTEGER PRIMARY KEY AUTOINCREMENT,"  \
               "LENGTH REAL, "                                 \
               "NAME TEXT"                                     \
               ");");

   return 0;
}


//***************************************************************************
// Show Database
//***************************************************************************

int LinslotWindow::showDb()
{
   db->execute("SELECT r.DRIVER1, r.DRIVER2, r.DATE, l.LAP_TIME from races as r join laps as l where r.RACE_ID = 1 and l.RACE_ID = r.RACE_ID;");
   db->show();

   return 0;
}

//***************************************************************************
// Store Race in Database
//***************************************************************************

int LinslotWindow::saveRace()
{
   sqlite3_stmt* sqlInsertRace;
   sqlite3_stmt* sqlInsertLap;
   int status;
   const char* duration;
   int driver1, driver2;
   int courseId;

   if (!*theSlots[0].driver || !*theSlots[1].driver)
   {
      QMessageBox::critical(this, "Fehler", "Kein Fahrer gewhlt!");
      return 0;
   }

   driver1 = getDriverId(theSlots[0].driver);
   driver2 = getDriverId(theSlots[1].driver);
   courseId = getCourseId();

   db->prepare("INSERT INTO races(DRIVER1,DRIVER2,DATE,LAPS,LAP_LENGTH,COURSE) "\
               "VALUES(?,?,datetime(?, 'unixepoch', 'utc'),?,?,?);", 
               sqlInsertRace);

   db->prepare("INSERT INTO laps(RACE_ID,DRIVER_NR,LAP_NR,LAP_TIME) "\
               "VALUES(?,?,?,?);", 
               sqlInsertLap);

   db->execute("BEGIN;");

   db->reset(sqlInsertRace);
   
   db->bindInt(sqlInsertRace,    1, driver1);
   db->bindInt(sqlInsertRace,    2, driver2);
   db->bindInt(sqlInsertRace,    3, time(0));
   db->bindInt(sqlInsertRace,    4, setupDialog->getRoundCount());
   db->bindDouble(sqlInsertRace, 5, setupDialog->getSlotLength());
   db->bindText(sqlInsertRace,   6, setupDialog->getCourseName());
   
   status = db->step(sqlInsertRace);
   
   if (status != 0)
      tell(eloAlways, "sqlite3_step(INSERT INTO races): '%s' (%d)",
           db->lastError(), status);
   
   int raceId = db->getInsertRowId();

   for (int l = 0; l < tableWidgetSlot1->rowCount(); l++)
   {
      db->reset(sqlInsertLap);

      duration = 0;

      if (tableWidgetSlot1->item(l, 1))
         duration = tableWidgetSlot1->item(l, 0)->text().toAscii();

      db->bindInt(sqlInsertLap, 1, raceId);       // RACE_ID
      db->bindInt(sqlInsertLap, 2, driver1);      // DRIVER_NR
      db->bindInt(sqlInsertLap, 3, l+1);          // LAP_NR

      if (duration)
         db->bindText(sqlInsertLap, 4, duration); // LAP_TIME
      else
         db->bindNull(sqlInsertLap, 4);           // LAP_TIME
      
      status = db->step(sqlInsertLap);
      
      if (status != 0)
         fprintf(stderr, "sqlite3_step(INSERT INTO laps): %s (%d)\n", 
                 db->lastError(), status);
   }

   for (int l = 0; l < tableWidgetSlot2->rowCount(); l++)
   {
      db->reset(sqlInsertLap);
      
      duration = 0;

      if (tableWidgetSlot2->item(l, 1))
         duration = tableWidgetSlot2->item(l, 0)->text().toAscii();

      db->bindInt(sqlInsertLap, 1, raceId);       // RACE_ID
      db->bindInt(sqlInsertLap, 2, driver2);      // DRIVER_NR
      db->bindInt(sqlInsertLap, 3, l+1);          // LAP_NR
      
      if (duration)
         db->bindText(sqlInsertLap, 4, duration); // LAP_TIME
      else
         db->bindNull(sqlInsertLap, 4);           // LAP_TIME

      status = db->step(sqlInsertLap);
      
      if (status != 0)
         fprintf(stderr, "sqlite3_step(INSERT INTO laps): %s (%d)\n", 
                 db->lastError(), status);
   }
   
   db->execute("COMMIT;");

   db->finalize(sqlInsertRace);
   db->finalize(sqlInsertLap);  

   return 0;
}

//***************************************************************************
// Get Driver Id
//***************************************************************************

int LinslotWindow::getDriverId(const char* driver)
{
   sqlite3_stmt* sqlSelectDriver;
   sqlite3_stmt* sqlInsertDriver;
   char* sql;
   int status;
   int id;

   // get driver ID's

   asprintf(&sql, "SELECT * from drivers where NAME=?;");
   db->prepare(sql, sqlSelectDriver);
   db->bindText(sqlSelectDriver, 1, driver);

   status = db->steps(sqlSelectDriver);

   db->reset(sqlSelectDriver);
   db->finalize(sqlSelectDriver);
   free(sql);  

   if (status == 0)
      id = db->getIntValueOf("DRIVER_ID");
   else
   {
      asprintf(&sql, "INSERT INTO drivers(NAME) values(?);");
      db->prepare(sql, sqlInsertDriver);
      db->execute("BEGIN;");
      db->bindText(sqlInsertDriver, 1, driver);
      db->step(sqlInsertDriver);
      db->execute("COMMIT;");
      id = db->getInsertRowId();
      db->finalize(sqlInsertDriver);
      free(sql);
   }

   return id;
}

//***************************************************************************
// Get Driver Id
//***************************************************************************

int LinslotWindow::getCourseId()
{
   sqlite3_stmt* sqlSelect;
   sqlite3_stmt* sqlInsert;
   char* sql;
   int status;
   int id;

   // get driver ID's

   asprintf(&sql, "SELECT * from cources where NAME=?;");
   db->prepare(sql, sqlSelect);
   db->bindText(sqlSelect, 1, setupDialog->getCourseName());

   status = db->steps(sqlSelect);

   db->reset(sqlSelect);
   db->finalize(sqlSelect);
   free(sql);  

   if (status == 0)
      id = db->getIntValueOf("COURSE_ID");
   else
   {
      asprintf(&sql, "INSERT INTO courses(NAME,LENGTH) values(?,?);");
      db->prepare(sql, sqlInsert);
      db->execute("BEGIN;");
      db->bindText(sqlInsert, 1, setupDialog->getCourseName());
      db->bindDouble(sqlInsert, 2, setupDialog->getSlotLength());
      db->step(sqlInsert);
      db->execute("COMMIT;");
      id = db->getInsertRowId();
      db->finalize(sqlInsert);
      free(sql);
   }

   return id;
}

//***************************************************************************
// Init Usb
//***************************************************************************

void LinslotWindow::on_pushButtonInitUsb_clicked()
{
   tell("reopening usb device");
   thread->reopen();
}
